@@ -1,6 +1,8 @@
-{-# OPTIONS_GHC -fno-warn-missing-methods -fno-warn-orphans #-}
+{-# OPTIONS_GHC -fno-warn-orphans #-}
+{-# LANGUAGE FunctionalDependencies #-}
 {-# LANGUAGE MultiParamTypeClasses #-}
-{-# LANGUAGE UndecidableInstances, UndecidableSuperClasses #-}
+{-# LANGUAGE UndecidableInstances #-}
+{-# LANGUAGE UndecidableSuperClasses #-}
 -- Copyright (C) 2010, 2011 Petr Rockai
 --
 -- Permission is hereby granted, free of charge, to any person
@@ -23,45 +25,43 @@
 -- CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 -- SOFTWARE.
 module Darcs.Patch.ApplyMonad
-  ( ApplyMonad(..), ApplyMonadTrans(..), ApplyMonadState(..)
-  , withFileNames, withFiles, ToTree(..)
+  ( ApplyMonad(..), ApplyMonadTrans(..), ApplyMonadOperations
+  , withFileNames
   , ApplyMonadTree(..)
+  , evalApplyMonad
   ) where
 
 import Darcs.Prelude
 
 import qualified Data.ByteString      as B
 import qualified Data.ByteString.Lazy as BL
-import qualified Data.Map             as M
 import qualified Darcs.Util.Tree.Monad as TM
+import Darcs.Patch.Object ( ObjectIdOf )
 import Darcs.Util.Tree ( Tree )
 import Data.Maybe ( fromMaybe )
 import Darcs.Util.Path ( AnchoredPath, movedirfilename, isPrefix )
+import Control.Monad.Catch ( MonadThrow(..) )
 import Control.Monad.State.Strict
-import Control.Monad.Identity( Identity )
-import Darcs.Patch.MonadProgress
+import Control.Monad.StrictIdentity (StrictIdentity(..) )
 
 import GHC.Exts ( Constraint )
 
-class ToTree s where
-  toTree :: s m -> Tree m
-
-instance ToTree Tree where
-  toTree = id
-
 class (Monad m, ApplyMonad state (ApplyMonadOver state m))
-      => ApplyMonadTrans (state :: (* -> *) -> *) m where
+      => ApplyMonadTrans state m where
   type ApplyMonadOver state m :: * -> *
   runApplyMonad :: (ApplyMonadOver state m) x -> state m -> m (x, state m)
 
-instance Monad m => ApplyMonadTrans Tree m where
+instance MonadThrow m => ApplyMonadTrans Tree m where
   type ApplyMonadOver Tree m = TM.TreeMonad m
   runApplyMonad = TM.virtualTreeMonad
 
-class ApplyMonadState (state :: (* -> *) -> *) where
-  type ApplyMonadStateOperations state :: (* -> *) -> Constraint
+evalApplyMonad
+  :: ApplyMonadTrans state m => ApplyMonadOver state m a -> state m -> m a
+evalApplyMonad action st = fst <$> runApplyMonad action st
 
-class Monad m => ApplyMonadTree m where
+type family ApplyMonadOperations (state :: (* -> *) -> *) :: (* -> *) -> Constraint
+
+class MonadThrow m => ApplyMonadTree m where
     -- a semantic, Tree-based interface for patch application
     mDoesDirectoryExist ::  AnchoredPath -> m Bool
     mDoesFileExist ::  AnchoredPath -> m Bool
@@ -69,46 +69,31 @@ class Monad m => ApplyMonadTree m where
     mCreateDirectory ::  AnchoredPath -> m ()
     mRemoveDirectory ::  AnchoredPath -> m ()
     mCreateFile ::  AnchoredPath -> m ()
-    mCreateFile f = mModifyFilePS f $ \_ -> return B.empty
     mRemoveFile ::  AnchoredPath -> m ()
     mRename ::  AnchoredPath -> AnchoredPath -> m ()
     mModifyFilePS ::  AnchoredPath -> (B.ByteString -> m B.ByteString) -> m ()
     mChangePref ::  String -> String -> String -> m ()
     mChangePref _ _ _ = return ()
 
-instance ApplyMonadState Tree where
-    type ApplyMonadStateOperations Tree = ApplyMonadTree
+type instance ApplyMonadOperations Tree = ApplyMonadTree
 
-class ( Monad m, Monad (ApplyMonadBase m)
-      , ApplyMonadStateOperations state m, ToTree state
+class ( Monad m
+      , ApplyMonadOperations state m
       )
-       -- ApplyMonadOver (ApplyMonadBase m) ~ m is *not* required in general,
-       -- since ApplyMonadBase is not injective
-       => ApplyMonad (state :: (* -> *) -> *) m where
-    type ApplyMonadBase m :: * -> *
+      => ApplyMonad (state :: (* -> *) -> *) m | m -> state where
 
-    nestedApply :: m x -> state (ApplyMonadBase m) -> m (x, state (ApplyMonadBase m))
-    liftApply :: (state (ApplyMonadBase m) -> (ApplyMonadBase m) x) -> state (ApplyMonadBase m)
-                 -> m (x, state (ApplyMonadBase m))
+    readFilePS :: ObjectIdOf state -> m B.ByteString
 
-    getApplyState :: m (state (ApplyMonadBase m))
-
-instance Monad m => ApplyMonad Tree (TM.TreeMonad m) where
-    type ApplyMonadBase (TM.TreeMonad m) = m
-    getApplyState = gets TM.tree
-    nestedApply a start = lift $ runApplyMonad a start
-    liftApply a start = do x <- gets TM.tree
-                           lift $ runApplyMonad (lift $ a x) start
-
-instance Monad m => ApplyMonadTree (TM.TreeMonad m) where
+instance MonadThrow m => ApplyMonad Tree (TM.TreeMonad m) where
+    readFilePS path = mReadFilePS path
 
+instance MonadThrow m => ApplyMonadTree (TM.TreeMonad m) where
     mDoesDirectoryExist p = TM.directoryExists p
     mDoesFileExist p = TM.fileExists p
-    mReadFilePS p = B.concat `fmap` BL.toChunks `fmap` TM.readFile p
-    mModifyFilePS p j = do have <- TM.fileExists p
-                           x <- if have then B.concat `fmap` BL.toChunks `fmap` TM.readFile p
-                                        else return B.empty
-                           TM.writeFile p . BL.fromChunks . (:[]) =<< j x
+    mReadFilePS p = BL.toStrict <$> TM.readFile p
+    mModifyFilePS p j =
+      TM.writeFile p . BL.fromStrict =<< j . BL.toStrict =<< TM.readFile p
+    mCreateFile p = TM.writeFile p BL.empty
     mCreateDirectory p = TM.createDirectory p
     mRename from to = TM.rename from to
     mRemoveDirectory = TM.unlink
@@ -118,7 +103,18 @@ instance Monad m => ApplyMonadTree (TM.TreeMonad m) where
 type OrigFileNameOf = (AnchoredPath, AnchoredPath)
 -- Touched files, new file list (after removes etc.) and rename details
 type FilePathMonadState = ([AnchoredPath], [AnchoredPath], [OrigFileNameOf])
-type FilePathMonad = State FilePathMonadState
+type FilePathMonad = StateT FilePathMonadState Pure
+
+newtype Pure a = Pure (StrictIdentity a)
+  deriving (Functor, Applicative, Monad)
+
+runPure :: Pure a -> a
+runPure (Pure (StrictIdentity x)) = x
+
+-- With "Data.Functor.Identity" this instance would not satisfy the law
+-- @throwM e >> x = throwM e@, which is why we use 'StrictIdentity'.
+instance MonadThrow Pure where
+  throwM e = Pure (error (show e))
 
 -- |trackOrigRename takes an old and new name and attempts to apply the mapping
 -- to the OrigFileNameOf pair. If the old name is the most up-to-date name of
@@ -136,17 +132,17 @@ trackOrigRename old new pair@(latest, from)
 -- present, a new list is generated from the filesnames.
 withFileNames :: Maybe [OrigFileNameOf] -> [AnchoredPath] -> FilePathMonad a
     -> FilePathMonadState
-withFileNames mbofnos fps x = execState x ([], fps, ofnos) where
+withFileNames mbofnos fps x = runPure $ execStateT x ([], fps, ofnos)
+  where
     ofnos = fromMaybe (map (\y -> (y, y)) fps) mbofnos
 
 instance ApplyMonad Tree FilePathMonad where
-    type ApplyMonadBase FilePathMonad = Identity
-
+    readFilePS = error "readFilePS not defined for FilePathMonad"
 
 instance ApplyMonadTree FilePathMonad where
     -- We can't check it actually is a directory here
     mDoesDirectoryExist p = gets $ \(_, fs, _) -> p `elem` fs
-
+    mDoesFileExist = mDoesDirectoryExist
     mCreateDirectory = mCreateFile
     mCreateFile f = modify $ \(ms, fs, rns) -> (f : ms, fs, rns)
     mRemoveFile f = modify $ \(ms, fs, rns) -> (f : ms, filter (/= f) fs, rns)
@@ -156,30 +152,4 @@ instance ApplyMonadTree FilePathMonad where
                                    , map (movedirfilename a b) fs
                                    , map (trackOrigRename a b) rns)
     mModifyFilePS f _ = mCreateFile f
-
-instance MonadProgress FilePathMonad where
-  runProgressActions = silentlyRunProgressActions
-
-type RestrictedApply = State (M.Map AnchoredPath B.ByteString)
-
-instance ApplyMonad Tree RestrictedApply where
-  type ApplyMonadBase RestrictedApply = Identity
-
-instance ApplyMonadTree RestrictedApply where
-  mDoesDirectoryExist _ = return True
-  mCreateDirectory _ = return ()
-  mRemoveFile f = modify $ M.delete f
-  mRemoveDirectory _ = return ()
-  mRename a b = modify $ M.mapKeys (movedirfilename a b)
-  mModifyFilePS f j = do look <- gets $ M.lookup f
-                         case look of
-                           Nothing -> return ()
-                           Just bits -> do
-                             new <- j bits
-                             modify $ M.insert f new
-
-instance MonadProgress RestrictedApply where
-  runProgressActions = silentlyRunProgressActions
-
-withFiles :: [(AnchoredPath, B.ByteString)] -> RestrictedApply a -> [(AnchoredPath, B.ByteString)]
-withFiles p x = M.toList $ execState x $ M.fromList p
+    mReadFilePS = error "mReadFilePS not defined for FilePathMonad"
