@@ -34,36 +34,31 @@ import System.IO.Error
     )
 
 import Darcs.Util.Tree( Tree )
-import Darcs.Util.External ( backupByCopying )
+import Darcs.Util.File ( backupByCopying )
 
 import Darcs.Patch
-    ( RepoPatch, IsRepoType, PrimOf, merge
+    ( RepoPatch, PrimOf, merge
     , effect
     , listConflictedFiles )
 import Darcs.Patch.Apply ( ApplyState )
-import Darcs.Patch.Ident ( merge2FL )
+import Darcs.Patch.Depends ( slightlyOptimizePatchset )
+import Darcs.Patch.Invertible ( mkInvertible )
 import Darcs.Patch.Named ( patchcontents, anonymous )
 import Darcs.Patch.PatchInfoAnd ( PatchInfoAnd, n2pia, hopefully )
-import Darcs.Patch.Progress( progressFL )
-import Darcs.Patch.Set ( PatchSet, Origin, patchSet2RL )
+import Darcs.Patch.Progress( progressFL, progressRL )
+import Darcs.Patch.Set ( PatchSet, Origin, appendPSFL, patchSet2RL )
 import Darcs.Patch.Witnesses.Ordered
     ( FL(..), RL(..), Fork(..), (:\/:)(..), (:/\:)(..), (+>+), (+<<+)
-    , mapFL_FL, concatFL, reverseFL )
+    , lengthFL, mapFL_FL, concatFL, reverseFL )
 import Darcs.Patch.Witnesses.Sealed ( Sealed(Sealed), seal )
 
 import Darcs.Repository.Flags
-    ( UseIndex
-    , ScanKnown
+    ( DiffOpts (..)
     , AllowConflicts (..)
+    , ResolveConflicts (..)
     , Reorder (..)
     , UpdatePending (..)
-    , ExternalMerge (..)
-    , Verbosity (..)
-    , Compression (..)
     , WantGuiPause (..)
-    , DiffAlgorithm (..)
-    , LookForMoves(..)
-    , LookForReplaces(..)
     )
 import Darcs.Repository.Hashed
     ( tentativelyAddPatches_
@@ -72,15 +67,15 @@ import Darcs.Repository.Hashed
     )
 import Darcs.Repository.Pristine
     ( applyToTentativePristine
-    , ApplyDir(..)
     )
-import Darcs.Repository.InternalTypes ( Repository, repoLocation )
+import Darcs.Repository.InternalTypes ( AccessType(RW), Repository, repoLocation )
 import Darcs.Repository.Pending ( setTentativePending )
 import Darcs.Repository.Resolution
-    ( externalResolution
-    , standardResolution
-    , StandardResolution(..)
+    ( StandardResolution(..)
     , announceConflicts
+    , externalResolution
+    , patchsetConflictResolutions
+    , standardResolution
     )
 import Darcs.Repository.State ( unrecordedChanges, readUnrecorded )
 
@@ -179,7 +174,7 @@ the pending patch, which is done when we finalize the tentative changes.
 
 Up to now we did not consider conflicts. Any new conflicts arising from the
 merges we made so far must be "resolved", that is, marked for manual
-resolution, if possible, or at least reported o the user. We made two
+resolution, if possible, or at least reported to the user. We made two
 merges, one with us and one with pw. It is important now to realize that our
 existing repo, and in particular the sequence us, could already be
 conflicted. Our job is to resolve only /new/ conflicts and not any
@@ -199,7 +194,7 @@ state U'':
          |
          U'
         / \
-    pw'  them''
+      pw' them''
       /     \
      T       U
     / \     /
@@ -224,52 +219,58 @@ The new pending patch starts out at the new tentative state, so as explained
 above, we set it to pw'+>+res, and again rely on sifting to commute out and
 drop anything we don't need.
 
-TODO: We should return a properly coerced @Repository rt p wR wU wT@.
+TODO: We should return a properly coerced @Repository 'RW p wU wR@.
 -}
 
-tentativelyMergePatches_ :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
+tentativelyMergePatches_ :: (RepoPatch p, ApplyState p ~ Tree)
                          => MakeChanges
-                         -> Repository rt p wR wU wR -> String
+                         -> Repository 'RW p wU wR -> String
                          -> AllowConflicts
-                         -> ExternalMerge -> WantGuiPause
-                         -> Compression -> Verbosity -> Reorder
-                         -> ( UseIndex, ScanKnown, DiffAlgorithm )
-                         -> Fork (PatchSet rt p)
-                                 (FL (PatchInfoAnd rt p))
-                                 (FL (PatchInfoAnd rt p)) Origin wR wY
+                         -> WantGuiPause
+                         -> Reorder
+                         -> DiffOpts
+                         -> Fork (PatchSet p)
+                                 (FL (PatchInfoAnd p))
+                                 (FL (PatchInfoAnd p)) Origin wR wY
                          -> IO (Sealed (FL (PrimOf p) wU))
-tentativelyMergePatches_ mc _repo cmd allowConflicts externalMerge wantGuiPause
-  compression verbosity reorder diffingOpts@(useidx, _, dflag) (Fork context us them) = do
-    (them' :/\: us')
-         <- return $ merge2FL (progressFL "Merging us" us)
-                              (progressFL "Merging them" them)
-    pw <- unrecordedChanges diffingOpts NoLookForMoves NoLookForReplaces _repo Nothing
+tentativelyMergePatches_ mc _repo cmd allowConflicts wantGuiPause
+  reorder diffingOpts@DiffOpts{..} (Fork context us them) = do
+    (them' :/\: us') <-
+      return $ merge (progressFL "Merging us" us :\/: progressFL "Merging them" them)
+    pw <- unrecordedChanges diffingOpts _repo Nothing
     -- Note: we use anonymous here to wrap the unrecorded changes.
     -- This is benign because we only retain the effect of the results
     -- of the merge (pw' and them'').
     anonpw <- n2pia `fmap` anonymous pw
     pw' :/\: them'' <- return $ merge (them' :\/: anonpw :>: NilFL)
-    let them''content = concatFL $ progressFL "Examining patches for conflicts" $
-                                mapFL_FL (patchcontents . hopefully) them''
-    let conflicts =
-          standardResolution
-            (patchSet2RL context +<<+ us :<: anonpw)
-            (reverseFL them'')
-    let standard_resolution = mangled conflicts
+    let them''content = concatFL $ mapFL_FL (patchcontents . hopefully) them''
+        no_conflicts_in_them =
+          null $ conflictedPaths $ patchsetConflictResolutions $
+          slightlyOptimizePatchset (appendPSFL context them)
+        conflicts =
+          let us'' = us' +>+ pw' in
+          -- This optimization is valid only if @them@ didn't have
+          -- (unresolved) conflicts in the first place
+          if lengthFL us'' < lengthFL them'' && no_conflicts_in_them then
+            standardResolution
+              (patchSet2RL context +<<+ them)
+              (progressRL "Examining patches for conflicts" $ reverseFL us'')
+          else
+            standardResolution
+              (patchSet2RL context +<<+ us :<: anonpw)
+              (progressRL "Examining patches for conflicts" $ reverseFL them'')
 
     debugMessage "Checking for conflicts..."
-    when (allowConflicts == YesAllowConflictsAndMark) $
+    when (allowConflicts == YesAllowConflicts MarkConflicts) $
         mapM_ backupByCopying $
         map (anchorPath (repoLocation _repo)) $
         conflictedPaths conflicts
 
     debugMessage "Announcing conflicts..."
-    have_conflicts <-
-        announceConflicts cmd allowConflicts externalMerge conflicts
+    have_conflicts <- announceConflicts cmd allowConflicts conflicts
 
     debugMessage "Checking for unrecorded conflicts..."
-    let pw'content = concatFL $ progressFL "Examining patches for conflicts" $
-                                mapFL_FL (patchcontents . hopefully) pw'
+    let pw'content = concatFL $ mapFL_FL (patchcontents . hopefully) pw'
     case listConflictedFiles pw'content of
         [] -> return ()
         fs -> do
@@ -286,62 +287,62 @@ tentativelyMergePatches_ mc _repo cmd allowConflicts externalMerge wantGuiPause
             exitSuccess
 
     debugMessage "Reading working tree..."
-    working <- readUnrecorded _repo useidx Nothing
+    working <- readUnrecorded _repo withIndex Nothing
 
     debugMessage "Working out conflict markup..."
     Sealed resolution <-
-        case (externalMerge , have_conflicts) of
-            (NoExternalMerge, _)       -> return $ if allowConflicts == YesAllowConflicts
-                                                     then seal NilFL
-                                                     else standard_resolution
-            (_, False)                 -> return $ standard_resolution
-            (YesExternalMerge c, True) -> externalResolution dflag working c wantGuiPause
-                                             (effect us +>+ pw) (effect them) them''content
+      if have_conflicts then
+        case allowConflicts of
+          YesAllowConflicts (ExternalMerge merge_cmd) ->
+            externalResolution diffAlg working merge_cmd wantGuiPause
+              (effect us +>+ pw) (effect them) them''content
+          YesAllowConflicts NoResolveConflicts -> return $ seal NilFL
+          YesAllowConflicts MarkConflicts -> return $ mangled conflicts
+          NoAllowConflicts -> error "impossible" -- was handled in announceConflicts
+      else return $ seal NilFL
 
     debugMessage "Adding patches to the inventory and writing new pending..."
     when (mc == MakeChanges) $ do
-        applyToTentativePristine _repo ApplyNormal verbosity them'
+        applyToTentativePristine _repo $ mkInvertible $
+          progressFL "Applying patches to pristine" them'
         -- these two cases result in the same trees (that's the idea of
         -- merging), so we only operate on the set of patches and do the
         -- adaption of pristine and pending in the common code below
         _repo <- case reorder of
             NoReorder -> do
-                tentativelyAddPatches_ DontUpdatePristine _repo
-                    compression verbosity NoUpdatePending them'
+                tentativelyAddPatches_ DontUpdatePristine _repo NoUpdatePending them'
             Reorder -> do
                 -- we do not actually remove any effect in the end, so
                 -- it would be wrong to update the unrevert bundle or
                 -- the working tree or pending
-                r1 <- tentativelyRemovePatches_ DontUpdatePristineNorRevert _repo
-                          compression NoUpdatePending us
-                r2 <- tentativelyAddPatches_ DontUpdatePristine r1
-                          compression verbosity NoUpdatePending them
-                tentativelyAddPatches_ DontUpdatePristine r2
-                    compression verbosity NoUpdatePending us'
+                _repo <- tentativelyRemovePatches_ DontUpdatePristineNorRevert _repo
+                          NoUpdatePending us
+                _repo <- tentativelyAddPatches_ DontUpdatePristine _repo
+                          NoUpdatePending them
+                tentativelyAddPatches_ DontUpdatePristine _repo NoUpdatePending us'
         setTentativePending _repo (effect pw' +>+ resolution)
     return $ seal (effect them''content +>+ resolution)
 
-tentativelyMergePatches :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
-                        => Repository rt p wR wU wR -> String
+tentativelyMergePatches :: (RepoPatch p, ApplyState p ~ Tree)
+                        => Repository 'RW p wU wR -> String
                         -> AllowConflicts
-                        -> ExternalMerge -> WantGuiPause
-                        -> Compression -> Verbosity -> Reorder
-                        -> ( UseIndex, ScanKnown, DiffAlgorithm )
-                        -> Fork (PatchSet rt p)
-                                (FL (PatchInfoAnd rt p))
-                                (FL (PatchInfoAnd rt p)) Origin wR wY
+                        -> WantGuiPause
+                        -> Reorder
+                        -> DiffOpts
+                        -> Fork (PatchSet p)
+                                (FL (PatchInfoAnd p))
+                                (FL (PatchInfoAnd p)) Origin wR wY
                         -> IO (Sealed (FL (PrimOf p) wU))
 tentativelyMergePatches = tentativelyMergePatches_ MakeChanges
 
-
-considerMergeToWorking :: (IsRepoType rt, RepoPatch p, ApplyState p ~ Tree)
-                       => Repository rt p wR wU wR -> String
+considerMergeToWorking :: (RepoPatch p, ApplyState p ~ Tree)
+                       => Repository 'RW p wU wR -> String
                        -> AllowConflicts
-                       -> ExternalMerge -> WantGuiPause
-                       -> Compression -> Verbosity -> Reorder
-                       -> ( UseIndex, ScanKnown, DiffAlgorithm )
-                       -> Fork (PatchSet rt p)
-                               (FL (PatchInfoAnd rt p))
-                               (FL (PatchInfoAnd rt p)) Origin wR wY
+                       -> WantGuiPause
+                       -> Reorder
+                       -> DiffOpts
+                       -> Fork (PatchSet p)
+                               (FL (PatchInfoAnd p))
+                               (FL (PatchInfoAnd p)) Origin wR wY
                        -> IO (Sealed (FL (PrimOf p) wU))
 considerMergeToWorking = tentativelyMergePatches_ DontMakeChanges
