@@ -18,6 +18,7 @@ should attempt to automatically create the patch-index.
 
 See <http://darcs.net/Internals/PatchIndex> for more information.
 -}
+{-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
 module Darcs.Repository.PatchIndex
     ( doesPatchIndexExist
     , isPatchIndexDisabled
@@ -37,17 +38,19 @@ module Darcs.Repository.PatchIndex
 import Darcs.Prelude
 
 import Control.Exception ( catch )
-import Control.Monad ( forM_, unless, when )
+import Control.Monad ( forM_, unless, when, (>=>) )
 import Control.Monad.State.Strict ( evalState, execState, State, gets, modify )
 
 import Data.Binary ( Binary, encodeFile, decodeFileOrFail )
 import qualified Data.ByteString as B
 import Data.Int ( Int8 )
-import Data.List ( group, mapAccumL, sort, nub, (\\) )
-import Data.Maybe ( fromJust, fromMaybe, isJust )
+import Data.List ( mapAccumL, sort, nub, (\\) )
+import Data.Maybe ( catMaybes, fromJust, fromMaybe )
+import qualified Data.IntSet as I
 import qualified Data.Map as M
 import qualified Data.Set as S
-import Data.Word ( Word32 )
+
+import Safe ( tailErr )
 
 import System.Directory
     ( createDirectory
@@ -63,7 +66,15 @@ import System.IO ( openFile, IOMode(WriteMode), hClose )
 import Darcs.Patch ( RepoPatch, listTouchedFiles )
 import Darcs.Patch.Apply ( ApplyState(..) )
 import Darcs.Patch.Index.Types
-import Darcs.Patch.Index.Monad ( applyToFileMods, makePatchID )
+    ( FileId(..)
+    , PatchId
+    , makePatchID
+    , pid2string
+    , short
+    , showFileId
+    , zero
+    )
+import Darcs.Patch.Index.Monad ( FileMod(..), applyToFileMods )
 import Darcs.Patch.Inspect ( PatchInspect )
 import Darcs.Patch.PatchInfoAnd ( PatchInfoAnd, info )
 import Darcs.Patch.Progress (progressFL )
@@ -81,17 +92,19 @@ import Darcs.Patch.Witnesses.Unsafe ( unsafeCoerceP, unsafeCoercePEnd )
 
 import Darcs.Repository.Format ( formatHas, RepoProperty( HashedInventory ) )
 import Darcs.Repository.InternalTypes ( Repository, repoLocation, repoFormat )
+import Darcs.Repository.Paths ( hashedInventoryPath )
 
 import Darcs.Util.Global ( darcsdir )
 import Darcs.Util.Hash ( sha256sum, showAsHex )
 import Darcs.Util.Lock ( withPermDir )
-import Darcs.Util.Path ( AnchoredPath, displayPath, toFilePath, isPrefix )
+import Darcs.Util.Path ( AnchoredPath, displayPath, isRoot, parents, toFilePath )
 import Darcs.Util.Progress ( debugMessage )
 import Darcs.Util.SignalHandler ( catchInterrupt )
 import Darcs.Util.Tree ( Tree(..) )
 
 type Map = M.Map
 type Set = S.Set
+type IntSet = I.IntSet
 
 data FileIdSpan = FidSpan
   !FileId                   -- ^ the fileid has some fixed name in the
@@ -105,10 +118,10 @@ data FilePathSpan = FpSpan
   !(Maybe PatchId)          -- ^ and (maybe) ending here
   deriving (Show, Eq, Ord)
 
--- | info about a given fileid, e.g.. is a file or a directory
+-- | info about a given fileid
 data FileInfo = FileInfo
-  { isFile :: Bool
-  , touching :: Set Word32  -- ^ first word of patch hash
+  { isFile :: Bool          -- ^ whether file or dir
+  , touching :: IntSet      -- ^ first words of patch hashes
   } deriving (Show, Eq, Ord)
 
 -- | timespans where a certain filename corresponds to a file with a given id
@@ -125,7 +138,7 @@ data PatchIndex = PatchIndex
   { pids :: [PatchId]
     -- ^ all the 'PatchId's tracked by this patch index, with the most
     -- recent patch at the head of the list (note, stored in the
-    -- reverse order to this on disk for backwards compatibility
+    -- reverse order on disk for backwards compatibility
     -- with an older format).
   , fidspans :: FileIdSpans
   , fpspans :: FilePathSpans
@@ -137,53 +150,62 @@ data PatchIndex = PatchIndex
 --           2 changes the pids order to newer-to-older
 --           3 changes FileName to AnchoredPath everywhere, which has
 --             different Binary (and Ord) instances
+--           4 adds all parent dirs of each file or dir as
+--             being touched by a patch
+--           5 replaces Set Word32 with IntSet
+
 version :: Int8
-version = 3
+version = 5
 
 type PIM a = State PatchIndex a
 
 -- | 'applyPatchMods pmods pindex' applies a list of PatchMods to the given
 --   patch index pindex
-applyPatchMods :: [(PatchId, [PatchMod AnchoredPath])] -> PatchIndex -> PatchIndex
+applyPatchMods :: [(PatchId, [FileMod AnchoredPath])] -> PatchIndex -> PatchIndex
 applyPatchMods pmods pindex =
   flip execState pindex $ mapM_ goList pmods
- where goList :: (PatchId, [PatchMod AnchoredPath]) -> PIM ()
+ where goList :: (PatchId, [FileMod AnchoredPath]) -> PIM ()
        goList (pid, mods) = do
            modify (\pind -> pind{pids = pid:pids pind})
-           mapM_ (curry go pid) (nubSeq mods)
-       -- nubSeq handles invalid patch in darcs repo:
-       --   move with identical target name "rename darcs_patcher to darcs-patcher."
-       nubSeq = map head . group
-       go :: (PatchId, PatchMod AnchoredPath) -> PIM ()
+           mapM_ (curry go pid) mods
+       go :: (PatchId, FileMod AnchoredPath) -> PIM ()
        go (pid, PCreateFile fn) = do
          fid <- createFidStartSpan fn pid
          startFpSpan fid fn pid
          createInfo fid True
-         insertTouch fid pid
+         insertTouch pid fid
+         insertParentsTouch pid fn
        go (pid, PCreateDir fn) = do
          fid <- createFidStartSpan fn pid
          startFpSpan fid fn pid
          createInfo fid False
-         insertTouch fid pid
+         insertTouch pid fid
+         insertParentsTouch pid fn
        go (pid, PTouch fn) = do
          fid <- lookupFid fn
-         insertTouch fid pid
+         insertTouch pid fid
+         insertParentsTouch pid fn
        go (pid, PRename oldfn newfn) = do
          fid <- lookupFid oldfn
          stopFpSpan fid pid
          startFpSpan fid newfn pid
-         insertTouch fid pid
+         insertTouch pid fid
+         insertParentsTouch pid oldfn
+         insertParentsTouch pid newfn
          stopFidSpan oldfn pid
          startFidSpan newfn pid fid
        go (pid, PRemove fn) = do
          fid <- lookupFid fn
-         insertTouch fid pid
+         insertTouch pid fid
+         insertParentsTouch pid fn
          stopFidSpan fn pid
          stopFpSpan fid pid
        go (pid, PDuplicateTouch fn) = do
          fidm <- gets fidspans
          case M.lookup fn fidm of
-           Just (FidSpan fid _ _:_) -> insertTouch fid pid
+           Just (FidSpan fid _ _:_) -> do
+             insertTouch pid fid
+             insertParentsTouch pid fn
            Nothing -> return ()
            Just [] -> error $ "applyPatchMods: impossible, no entry for "++show fn
                               ++" in FileIdSpans in duplicate, empty list"
@@ -235,14 +257,20 @@ stopFidSpan fn pend = modify (\pind -> pind {fidspans=M.alter alt fn (fidspans p
 -- | insert touching patchid for given file id
 createInfo :: FileId -> Bool -> PIM ()
 createInfo fid isF = modify (\pind -> pind {infom=M.alter alt fid (infom pind)})
-  where alt Nothing = Just (FileInfo isF S.empty)
-        alt (Just _) = Just (FileInfo isF S.empty) -- forget old false positives
+  where alt Nothing = Just (FileInfo isF I.empty)
+        alt (Just _) = Just (FileInfo isF I.empty) -- forget old false positives
 
 -- | insert touching patchid for given file id
-insertTouch :: FileId -> PatchId -> PIM ()
-insertTouch fid pid = modify (\pind -> pind {infom=M.alter alt fid (infom pind)})
+insertTouch :: PatchId -> FileId -> PIM ()
+insertTouch pid fid = modify (\pind -> pind {infom=M.alter alt fid (infom pind)})
   where alt Nothing =  error "impossible: Fileid does not exist"
-        alt (Just (FileInfo isF pids)) = Just (FileInfo isF (S.insert (short pid) pids))
+        alt (Just (FileInfo isF pids)) = Just (FileInfo isF (I.insert (short pid) pids))
+
+-- | insert touching patchid for the parents of a given path
+insertParentsTouch :: PatchId -> AnchoredPath -> PIM ()
+insertParentsTouch pid path =
+  forM_ (filter (not . isRoot) (parents path)) $
+    lookupFid >=> insertTouch pid
 
 -- | lookup current fid of filepath
 lookupFid :: AnchoredPath -> PIM FileId
@@ -261,56 +289,20 @@ lookupFid' fn = do
     _ -> return Nothing
 
 
--- | lookup all the file ids of a given path
-lookupFidf' :: AnchoredPath -> PIM [FileId]
-lookupFidf' fn = do
-   fidm <- gets fidspans
-   case M.lookup fn fidm of
-      Just spans -> return $ map (\(FidSpan fid _ _) -> fid) spans
-      Nothing ->
-         error $ "lookupFidf': no entry for " ++ show fn ++ " in FileIdSpans"
-
--- |  return all fids of matching subpaths
---    of the given filepath
-lookupFids :: AnchoredPath -> PIM [FileId]
-lookupFids fn = do
-   fid_spans <- gets fidspans
-   file_idss <- mapM lookupFidf' $
-      filter (isPrefix fn) (fpSpans2filePaths' fid_spans)
-   return $ nub $ concat file_idss
-
--- | returns a single file id if the given path is a file
---   if it is a directory, if returns all the file ids of all paths inside it,
---   at any point in repository history
-lookupFids' :: AnchoredPath -> PIM [FileId]
-lookupFids' fn = do
-  info_map <- gets infom
-  fps_spans <- gets fpspans
-  a <- lookupFid' fn
-  if isJust a then do
-                let fid = fromJust a
-                case M.lookup fid info_map of
-                  Just (FileInfo True _) -> return [fid]
-                  Just (FileInfo False _) ->
-                    let file_names = map (\(FpSpan x _ _) -> x) (fps_spans M.! fid)
-                    in nub . concat <$> mapM lookupFids file_names
-                  Nothing -> error "lookupFids' : could not find file"
-              else return []
-
 -- | Creates patch index that corresponds to all patches in repo.
 createPatchIndexDisk
   :: (RepoPatch p, ApplyState p ~ Tree)
-  => Repository rt p wR wU wT
-  -> PatchSet rt p Origin wR
+  => Repository rt p wU wR
+  -> PatchSet p Origin wR
   -> IO ()
 createPatchIndexDisk repository ps = do
   let patches = mapFL Sealed2 $ progressFL "Create patch index" $ patchSet2FL ps
-  createPatchIndexFrom repository $ patches2patchMods patches S.empty
+  createPatchIndexFrom repository $ patches2fileMods patches S.empty
 
 -- | convert patches to patchmods
-patches2patchMods :: (Apply p, PatchInspect p, ApplyState p ~ Tree)
-                  => [Sealed2 (PatchInfoAnd rt p)] -> Set AnchoredPath -> [(PatchId, [PatchMod AnchoredPath])]
-patches2patchMods patches fns = snd $ mapAccumL go fns patches
+patches2fileMods :: (Apply p, PatchInspect p, ApplyState p ~ Tree)
+                  => [Sealed2 (PatchInfoAnd p)] -> Set AnchoredPath -> [(PatchId, [FileMod AnchoredPath])]
+patches2fileMods patches fns = snd $ mapAccumL go fns patches
   where
     go filenames (Sealed2 p) = (filenames', (pid, pmods_effect ++ pmods_dup))
       where pid = makePatchID . info $ p
@@ -367,8 +359,8 @@ removePidSuffix pid2idx oldpids@(oldpid:_) (PatchIndex pids fidspans fpspans inf
 -- | update the patch index to the current state of the repository
 updatePatchIndexDisk
     :: (RepoPatch p, ApplyState p ~ Tree)
-    => Repository rt p wR wU wT
-    -> PatchSet rt p Origin wR
+    => Repository rt p wU wR
+    -> PatchSet p Origin wR
     -> IO ()
 updatePatchIndexDisk repo patches = do
     let repodir = repoLocation repo
@@ -382,7 +374,7 @@ updatePatchIndexDisk repo patches = do
         cdir = repodir </> indexDir
     -- reread to prevent holding onto patches for too long
     let newpatches = drop len_common $ mapFL seal2 flpatches
-        newpmods = patches2patchMods newpatches filenames
+        newpmods = patches2fileMods newpatches filenames
     inv_hash <- getInventoryHash repodir
     storePatchIndex cdir inv_hash (applyPatchMods newpmods pindex')
   where
@@ -395,8 +387,8 @@ updatePatchIndexDisk repo patches = do
 
 -- | 'createPatchIndexFrom repo pmods' creates a patch index from the given
 --   patchmods.
-createPatchIndexFrom :: Repository rt p wR wU wT
-                     -> [(PatchId, [PatchMod AnchoredPath])] -> IO ()
+createPatchIndexFrom :: Repository rt p wU wR
+                     -> [(PatchId, [FileMod AnchoredPath])] -> IO ()
 createPatchIndexFrom repo pmods = do
     inv_hash <- getInventoryHash repodir
     storePatchIndex cdir inv_hash (applyPatchMods pmods emptyPatchIndex)
@@ -406,7 +398,7 @@ createPatchIndexFrom repo pmods = do
 
 getInventoryHash :: FilePath -> IO String
 getInventoryHash repodir = do
-  inv <- B.readFile (repodir </> darcsdir </> "hashed_inventory")
+  inv <- B.readFile (repodir </> hashedInventoryPath)
   return $ sha256sum inv
 
 -- | Load patch-index from disk along with some meta data.
@@ -423,15 +415,19 @@ loadPatchIndex repodir = do
 
 -- | If patch-index is useful as it is now, read it. If not, create or update it, then read it.
 loadSafePatchIndex :: (RepoPatch p, ApplyState p ~ Tree)
-                   => Repository rt p wR wU wT
-                   -> PatchSet rt p Origin wR     -- ^ PatchSet of the repository, used if we need to create the patch-index.
+                   => Repository rt p wU wR
+                   -> PatchSet p Origin wR     -- ^ PatchSet of the repository, used if we need to create the patch-index.
                    -> IO PatchIndex
 loadSafePatchIndex repo ps = do
    let repodir = repoLocation repo
    can_use <- isPatchIndexInSync repo
    (_,_,_,pi) <-
      if can_use
-       then loadPatchIndex repodir
+       then do
+          debugMessage "Loading patch index..."
+          r <- loadPatchIndex repodir
+          debugMessage "Done."
+          return r
        else do createOrUpdatePatchIndexDisk repo ps
                loadPatchIndex repodir
    return pi
@@ -462,14 +458,16 @@ isPatchIndexDisabled repodir = doesFileExist (repodir </> darcsdir  </> noPatchI
 --   2. if patch index exists, update it
 --   3. if not, create it from scratch
 createOrUpdatePatchIndexDisk :: (RepoPatch p, ApplyState p ~ Tree)
-                             => Repository rt p wR wU wT -> PatchSet rt p Origin wR -> IO ()
+                             => Repository rt p wU wR -> PatchSet p Origin wR -> IO ()
 createOrUpdatePatchIndexDisk repo ps = do
+   debugMessage "createOrUpdatePatchIndexDisk: start"
    let repodir = repoLocation repo
    removeFile (repodir </> darcsdir </> noPatchIndex) `catch` \(_ :: IOError) -> return ()
    dpie <- doesPatchIndexExist repodir
    if dpie
       then updatePatchIndexDisk repo ps
       else createPatchIndexDisk repo ps
+   debugMessage "createOrUpdatePatchIndexDisk: done"
 
 -- | Read-only. Checks the two following things:
 --
@@ -478,7 +476,7 @@ createOrUpdatePatchIndexDisk repo ps = do
 --
 -- Then only if it exists and it is not explicitely disabled, returns @True@, else returns @False@
 -- (or an error if it exists and is explicitely disabled at the same time).
-canUsePatchIndex :: Repository rt p wR wU wT -> IO Bool
+canUsePatchIndex :: Repository rt p wU wR -> IO Bool
 canUsePatchIndex repo = do
      let repodir = repoLocation repo
      piExists <- doesPatchIndexExist repodir
@@ -492,7 +490,7 @@ canUsePatchIndex repo = do
 -- | Creates patch-index (ignoring whether it is explicitely disabled).
 --   If it is ctrl-c'ed, then aborts, delete patch-index and mark it as disabled.
 createPIWithInterrupt :: (RepoPatch p, ApplyState p ~ Tree)
-                      => Repository rt p wR wU wT -> PatchSet rt p Origin wR -> IO ()
+                      => Repository rt p wU wR -> PatchSet p Origin wR -> IO ()
 createPIWithInterrupt repo ps = do
     let repodir = repoLocation repo
     putStrLn "Creating a patch index, please wait. To stop press Ctrl-C"
@@ -500,9 +498,9 @@ createPIWithInterrupt repo ps = do
       createPatchIndexDisk repo ps
       putStrLn "Created patch index.") `catchInterrupt` (putStrLn "Patch Index Disabled" >> deletePatchIndex repodir)
 
--- | Checks if patch-index exists and is in sync with repository (more precisely with @_darcs\/hashed_inventory@).
+-- | Checks if patch-index exists and is in sync with repository.
 --   That is, checks if patch-index can be used as it is now.
-isPatchIndexInSync :: Repository rt p wR wU wT -> IO Bool
+isPatchIndexInSync :: Repository rt p wU wR -> IO Bool
 isPatchIndexInSync repo = do
    let repodir = repoLocation repo
    dpie <- doesPatchIndexExist repodir
@@ -615,8 +613,8 @@ dumpFilePathSpans fpspans =
            | (fid, fns) <- M.toList fpspans, FpSpan fn from mto <- fns]
 
 dumpTouchingMap :: InfoMap -> String
-dumpTouchingMap infom = unlines [showFileId fid++(if isF then "" else "/")++" -> "++ showAsHex w32
-                                | (fid,FileInfo isF w32s) <- M.toList infom, w32 <- S.elems w32s]
+dumpTouchingMap infom = unlines [showFileId fid++(if isF then "" else "/")++" -> "++ showAsHex (fromIntegral i)
+                                | (fid,FileInfo isF w32s) <- M.toList infom, i <- I.elems w32s]
 
 -- | return set of current filepaths in patch index
 fpSpans2filePaths :: FilePathSpans -> InfoMap -> [FilePath]
@@ -624,14 +622,10 @@ fpSpans2filePaths fpSpans infom =
   sort [displayPath fn ++ (if isF then "" else "/") | (fid,FpSpan fn _ Nothing:_) <- M.toList fpSpans,
                                                 let Just (FileInfo isF _) = M.lookup fid infom]
 
--- | return set of current filepaths in patch index, for internal use
-fpSpans2filePaths' :: FileIdSpans -> [AnchoredPath]
-fpSpans2filePaths' fidSpans = [fp | (fp, _)  <- M.toList fidSpans]
-
 -- | Checks if patch index can be created and build it with interrupt.
 attemptCreatePatchIndex
   :: (RepoPatch p, ApplyState p ~ Tree)
-  => Repository rt p wR wU wT -> PatchSet rt p Origin wR -> IO ()
+  => Repository rt p wU wR -> PatchSet p Origin wR -> IO ()
 attemptCreatePatchIndex repo ps = do
   canCreate <- canCreatePI repo
   when canCreate $ createPIWithInterrupt repo ps
@@ -639,7 +633,7 @@ attemptCreatePatchIndex repo ps = do
 -- | Checks whether a patch index can (and should) be created. If we are not in
 -- an old-fashioned repo, and if we haven't been told not to, then we should
 -- create a patch index if it doesn't already exist.
-canCreatePI :: Repository rt p wR wU wT -> IO Bool
+canCreatePI :: Repository rt p wU wR -> IO Bool
 canCreatePI repo =
     (not . or) <$> sequence [ doesntHaveHashedInventory (repoFormat repo)
                             , isPatchIndexDisabled repodir
@@ -654,12 +648,12 @@ canCreatePI repo =
 -- disabled, silently create it. (Also, if it is out-of-sync, which should not
 -- happen, silently update it).
 getRelevantSubsequence
-    :: (RepoPatch p, ApplyState p ~ Tree, a ~ PatchInfoAnd rt p)
+    :: (RepoPatch p, ApplyState p ~ Tree, a ~ PatchInfoAnd p)
     => Sealed ((RL a) wK)
     -- ^ Sequence of patches you want to filter
-    -> Repository rt p wR wU wR
+    -> Repository rt p wU wR
     -- ^ The repository (to attempt loading patch-index from its path)
-    -> PatchSet rt p Origin wR
+    -> PatchSet p Origin wR
     -- ^ PatchSet of repository (in case we need to create patch-index)
     -> [AnchoredPath]
     -- ^ File(s) about which you want patches from given sequence
@@ -669,38 +663,39 @@ getRelevantSubsequence pxes repository ps fns = do
     pi@(PatchIndex _ _ _ infom) <- loadSafePatchIndex repository ps
     let fids = map (\fn -> evalState (lookupFid fn) pi) fns
         pidss = map ((\(FileInfo _ a) -> a) . fromJust . (`M.lookup` infom)) fids
-        pids = S.unions pidss
+        pids = I.unions pidss
     let flpxes = reverseRL $ unseal unsafeCoercePEnd pxes
     return . seal $ keepElems flpxes NilRL pids
   where
-    keepElems :: (RepoPatch p, ApplyState p ~ Tree, a ~ PatchInfoAnd rt p)
-              => FL a wX wY -> RL a wB wX -> S.Set Word32 -> RL a wP wQ
+    keepElems :: (RepoPatch p, ApplyState p ~ Tree, a ~ PatchInfoAnd p)
+              => FL a wX wY -> RL a wB wX -> IntSet -> RL a wP wQ
     keepElems NilFL acc _ = unsafeCoerceP acc
     keepElems (x :>: xs) acc pids
-      | short (makePatchID $ info x) `S.member` pids = keepElems xs (acc :<: x) pids
+      | short (makePatchID $ info x) `I.member` pids = keepElems xs (acc :<: x) pids
       | otherwise = keepElems (unsafeCoerceP xs) acc pids
 
-type PatchFilter rt p = [AnchoredPath] -> [Sealed2 (PatchInfoAnd rt p)] -> IO [Sealed2 (PatchInfoAnd rt p)]
+type PatchFilter p = [AnchoredPath] -> [Sealed2 (PatchInfoAnd p)] -> IO [Sealed2 (PatchInfoAnd p)]
 
--- | If a patch index is available, returns a filter that takes a list of files and returns
---   a @PatchFilter@ that only keeps patches that modify the given list of files.
---   If patch-index cannot be used, return the original input.
---   If patch-index does not exist and is not explicitely disabled, silently create it.
---   (Also, if it is out-of-sync, which should not happen, silently update it).
+-- | If a patch index is available, returns a filter that takes a list of files
+--   and returns a @PatchFilter@ that only keeps patches that modify the given
+--   list of files. If patch-index cannot be used, return the original input.
+--   If patch-index does not exist and is not explicitely disabled, silently
+--   create it. (Also, if it is out-of-sync, which should not happen, silently
+--   update it).
 maybeFilterPatches
     :: (RepoPatch p, ApplyState p ~ Tree)
-    => Repository rt p wR wU wT  -- ^ The repository
-    -> PatchSet rt p Origin wR   -- ^ PatchSet of patches of repository (in case patch-index needs to be created)
-    -> PatchFilter rt p          -- ^ PatchFilter ready to be used by SelectChanges.
+    => Repository rt p wU wR  -- ^ The repository
+    -> PatchSet p Origin wR   -- ^ PatchSet of patches of repository (in case patch-index needs to be created)
+    -> PatchFilter p          -- ^ PatchFilter ready to be used by SelectChanges.
 maybeFilterPatches repo ps fps ops = do
     usePI <- canUsePatchIndex repo
     if usePI
       then do
         pi@(PatchIndex _ _ _ infom) <- loadSafePatchIndex repo ps
-        let fids = concatMap ((\fn -> evalState (lookupFids' fn) pi)) fps
-            npids = S.unions $ map (touching.fromJust.(`M.lookup` infom)) fids
+        let fids = catMaybes $ map ((\fn -> evalState (lookupFid' fn) pi)) fps
+            npids = I.unions $ map (touching.fromJust.(`M.lookup` infom)) fids
         return $ filter
-          (flip S.member npids . (unseal2 (short . makePatchID . info))) ops
+          (flip I.member npids . (unseal2 (short . makePatchID . info))) ops
       else return ops
 
 -- | Dump information in patch index. Patch-index should be checked to exist beforehand. Read-only.
@@ -757,14 +752,14 @@ piTest repodir = do
       let f :: FilePathSpan -> FilePathSpan -> Bool
           f (FpSpan _ x _) (FpSpan _ _ (Just y)) = x == y
           f _ _ = error "adj test of fpspans fail"
-      unless (and $ zipWith f spans (tail spans)) (fail $ "Adjcency test failed! fid: " ++ show fid)
+      unless (and $ zipWith f spans (tailErr spans)) (fail $ "Adjcency test failed! fid: " ++ show fid)
    putStrLn "fpspans tests passed"
 
    -- test infomap
    putStrLn "infom"
    putStrLn "==========="
    putStrLn $ "Valid fid test: " ++ (show.and $ map (`M.member` fpspans) (M.keys infom))
-   putStrLn $ "Valid pid test: " ++ (show.flip S.isSubsetOf (S.fromList $ map short pids)  . S.unions . map touching . M.elems $ infom)
+   putStrLn $ "Valid pid test: " ++ (show.flip I.isSubsetOf (I.fromList $ map short pids)  . I.unions . map touching . M.elems $ infom)
    where
           isInOrder :: Eq a => [a] -> [a] -> Bool
           isInOrder (x:xs) (y:ys) | x == y = isInOrder xs ys
