@@ -16,6 +16,7 @@
 --  Boston, MA 02110-1301, USA.
 module Darcs.Repository.Resolution
     ( standardResolution
+    , rebaseResolution
     , externalResolution
     , patchsetConflictResolutions
     , StandardResolution(..)
@@ -37,18 +38,18 @@ import Control.Monad ( when )
 
 import Darcs.Repository.Diff( treeDiff )
 import Darcs.Patch
-    ( PrimOf
-    , PrimPatchBase
+    ( Named
+    , PrimOf
     , RepoPatch
     , applyToTree
     , effect
     , effectOnPaths
     , invert
     , listConflictedFiles
+    , patchcontents
     , resolveConflicts
     )
 import Darcs.Patch.Apply( ApplyState )
-import Darcs.Patch.Commute ( Commute )
 import Darcs.Patch.Conflict ( Conflict, ConflictDetails(..), Mangled, Unravelled )
 import Darcs.Patch.Inspect ( listTouchedFiles )
 import Darcs.Patch.Merge ( mergeList )
@@ -60,20 +61,20 @@ import Darcs.Util.Path
     , filterPaths
     , toFilePath
     )
-import Darcs.Patch.Witnesses.Ordered ( FL(..), RL(..) )
+import Darcs.Patch.Witnesses.Ordered ( FL(..), RL(..), concatRLFL, mapRL_RL )
 import Darcs.Patch.Witnesses.Sealed ( Sealed(..), unseal, unFreeLeft )
 
 import Darcs.Util.CommandLine ( parseCmd )
-import Darcs.Patch.PatchInfoAnd ( PatchInfoAnd )
+import Darcs.Patch.PatchInfoAnd ( PatchInfoAnd, hopefully )
 import Darcs.Util.Prompt ( askEnter )
 import Darcs.Patch.Set ( PatchSet(..), Origin, patchSet2RL )
 import Darcs.Repository.Prefs ( filetypeFunction )
 import Darcs.Util.Exec ( exec, Redirect(..) )
 import Darcs.Util.Lock ( withTempDir )
-import Darcs.Util.External ( cloneTree )
+import Darcs.Util.File ( copyTree )
 import Darcs.Repository.Flags
     ( AllowConflicts (..)
-    , ExternalMerge (..)
+    , ResolveConflicts (..)
     , WantGuiPause (..)
     , DiffAlgorithm (..)
     )
@@ -93,11 +94,29 @@ data StandardResolution prim wX =
     conflictedPaths :: [AnchoredPath]
   }
 
-standardResolution :: (Commute p, PrimPatchBase p, Conflict p)
-                   => RL (PatchInfoAnd rt p) wO wX
-                   -> RL (PatchInfoAnd rt p) wX wY
+standardResolution :: (RepoPatch p)
+                   => RL (PatchInfoAnd p) wO wX
+                   -> RL (PatchInfoAnd p) wX wY
                    -> StandardResolution (PrimOf p) wY
 standardResolution context interesting =
+  mangleConflicts $ resolveConflicts context interesting
+
+-- | Like 'standardResolution' but it doesn't use the @instance (Named p)@
+-- because the traling list of patches may contain "fake" conflictors.
+rebaseResolution
+  :: (Conflict p, PrimPatch (PrimOf p))
+  => RL (PatchInfoAnd p) wO wX
+  -> RL (Named p) wX wY
+  -> StandardResolution (PrimOf p) wY
+rebaseResolution context interesting =
+    mangleConflicts $ resolveConflicts context_patches interesting_patches
+  where
+    context_patches = concatRLFL (mapRL_RL (patchcontents . hopefully) context)
+    interesting_patches = concatRLFL (mapRL_RL patchcontents interesting)
+
+mangleConflicts
+  :: (PrimPatch prim) => [ConflictDetails prim wX] -> StandardResolution prim wX
+mangleConflicts conflicts =
   case mergeList $ catMaybes $ map conflictMangled conflicts of
     Right mangled -> StandardResolution {..}
     Left (Sealed ps, Sealed qs) ->
@@ -107,23 +126,27 @@ standardResolution context interesting =
         $$ redText "conflicts with"
         $$ displayPatch qs
   where
-    conflicts = resolveConflicts context interesting
     unmangled = map conflictParts $ filter (isNothing . conflictMangled) conflicts
     conflictedPaths =
       nubSort $
       concatMap (unseal listTouchedFiles) (concatMap conflictParts conflicts)
 
-warnUnmangled :: PrimPatch prim => StandardResolution prim wX -> IO ()
-warnUnmangled StandardResolution {..}
+warnUnmangled
+  :: PrimPatch prim => Maybe [AnchoredPath] -> StandardResolution prim wX -> IO ()
+warnUnmangled mpaths StandardResolution {..}
   | null unmangled = return ()
-  | otherwise = ePutDocLn $ showUnmangled unmangled
+  | otherwise = ePutDocLn $ showUnmangled mpaths unmangled
 
-showUnmangled :: PrimPatch prim => [Unravelled prim wX] -> Doc
-showUnmangled = vcat . map showUnmangledConflict
+showUnmangled
+  :: PrimPatch prim => Maybe [AnchoredPath] -> [Unravelled prim wX] -> Doc
+showUnmangled mpaths = vcat . map showUnmangledConflict . filter (affected mpaths)
   where
     showUnmangledConflict unravelled =
       redText "Cannot mark these conflicting patches:" $$
       showUnravelled (redText "versus") unravelled
+    affected Nothing _ = True
+    affected (Just paths) unravelled =
+      any (`elem` paths) $ concatMap (unseal listTouchedFiles) unravelled
 
 showUnravelled :: PrimPatch prim => Doc -> Unravelled prim wX -> Doc
 showUnravelled sep =
@@ -132,26 +155,27 @@ showUnravelled sep =
 announceConflicts :: PrimPatch prim
                   => String
                   -> AllowConflicts
-                  -> ExternalMerge
                   -> StandardResolution prim wX
                   -> IO Bool
-announceConflicts cmd allowConflicts externalMerge conflicts =
+announceConflicts cmd allowConflicts conflicts =
   case nubSort (conflictedPaths conflicts) of
     [] -> return False
     cfs -> do
       ePutDocLn $ vcat $ redText
         "We have conflicts in the following files:" : map (text . displayPath) cfs
-      when (allowConflicts == YesAllowConflictsAndMark) $ warnUnmangled conflicts
-      if allowConflicts `elem` [YesAllowConflicts,YesAllowConflictsAndMark]
-              || externalMerge /= NoExternalMerge
-        then return True
-        else fail $
+      case allowConflicts of
+        NoAllowConflicts ->
+          fail $
           "Refusing to "++cmd++" patches leading to conflicts.\n"++
           "If you would rather apply the patch and mark the conflicts,\n"++
           "use the --mark-conflicts or --allow-conflicts options to "++cmd++"\n"++
           "These can set as defaults by adding\n"++
           " "++cmd++" mark-conflicts\n"++
           "to "++darcsdir++"/prefs/defaults in the target repo. "
+        YesAllowConflicts MarkConflicts -> do
+          warnUnmangled Nothing conflicts
+          return True
+        _ -> return True
 
 externalResolution :: forall p wX wY wZ wA. (RepoPatch p, ApplyState p ~ Tree.Tree)
                    => DiffAlgorithm
@@ -160,15 +184,15 @@ externalResolution :: forall p wX wY wZ wA. (RepoPatch p, ApplyState p ~ Tree.Tr
                    -> WantGuiPause        -- ^ tell whether we want GUI pause
                    -> FL (PrimOf p) wX wY -- ^ our effect
                    -> FL (PrimOf p) wX wZ -- ^ their effect
-                   -> FL p wY wA          -- ^ them merged (standard_resolution)
+                   -> FL p wY wA          -- ^ them merged
                    -> IO (Sealed (FL (PrimOf p) wA))
 externalResolution diffa s1 c wantGuiPause p1 p2 pmerged = do
  sa <- applyToTree (invert p1) s1
  sm <- applyToTree pmerged s1
  s2 <- applyToTree p2 sa
  let nms = listConflictedFiles pmerged
-     nas = effectOnPaths (invert (effect pmerged)) nms
-     n1s = effectOnPaths p1 nas
+     n1s = effectOnPaths (invert (effect pmerged)) nms
+     nas = effectOnPaths (invert p1) n1s
      n2s = effectOnPaths p2 nas
      ns = zip4 (tofp nas) (tofp n1s) (tofp n2s) (tofp nms)
      tofp = map (anchorPath "")
@@ -189,7 +213,7 @@ externalResolution diffa s1 c wantGuiPause p1 p2 pmerged = do
          setCurrentDirectory former_dir
          withTempDir "cleanmerged" $ \absdc -> do
            let dc = toFilePath absdc
-           cloneTree dm "."
+           copyTree dm "."
            setCurrentDirectory former_dir
            withTempDir "version2" $ \absd2 -> do
              let d2 = toFilePath absd2
@@ -212,7 +236,7 @@ externallyResolveFile c wantGuiPause da d1 d2 dm (fa, f1, f2, fm) = do
     putStrLn $ "Merging file "++fm++" by hand."
     ec <- run c [('1', d1</>f1), ('2', d2</>f2), ('a', da</>fa), ('o', dm</>fm), ('%', "%")]
     when (ec /= ExitSuccess) $
-         putStrLn $ "External merge command exited with " ++ show ec
+         fail $ "External merge command exited with " ++ show ec
     when (wantGuiPause == YesWantGuiPause) $
         askEnter "Hit return to move on, ^C to abort the whole operation..."
 
@@ -223,11 +247,11 @@ run c replacements =
     Right (c2,_) -> rr c2
     where rr (command:args) = do putStrLn $ "Running command '" ++
                                             unwords (command:args) ++ "'"
-                                 exec command args (Null,Null,Null)
+                                 exec command args (Null,AsIs,AsIs)
           rr [] = return ExitSuccess
 
 patchsetConflictResolutions :: RepoPatch p
-                            => PatchSet rt p Origin wX
+                            => PatchSet p Origin wX
                             -> StandardResolution (PrimOf p) wX
 patchsetConflictResolutions (PatchSet ts xs) =
   -- optimization: all patches before the latest known clean tag
