@@ -19,7 +19,7 @@ module Darcs.UI.Commands.Remove ( remove, rm, unadd ) where
 
 import Darcs.Prelude
 
-import Control.Monad ( when, foldM )
+import Control.Monad ( when, foldM, void )
 import Darcs.UI.Commands
     ( DarcsCommand(..)
     , withStdOpts, nodefaults
@@ -29,17 +29,18 @@ import Darcs.UI.Commands
     )
 import Darcs.UI.Completion ( knownFileArgs )
 import Darcs.UI.Flags
-    ( DarcsFlag, useCache, dryRun, umask, diffAlgorithm, quiet, pathsFromArgs )
-import Darcs.UI.Options ( (^), odesc, ocheck, defaultFlags, parseFlags, (?) )
+    ( DarcsFlag, diffingOpts
+    , useCache, umask, diffAlgorithm, pathsFromArgs )
+import Darcs.UI.Options ( (^), parseFlags, (?) )
 import qualified Darcs.UI.Options.All as O
 
-import Darcs.Repository.Flags ( UpdatePending (..) )
 import Darcs.Repository
     ( Repository
     , withRepoLock
     , RepoJob(..)
     , addToPending
-    , readRecordedAndPending
+    , finalizeRepositoryChanges
+    , readPristineAndPending
     , readUnrecorded
     )
 import Darcs.Repository.Diff( treeDiff )
@@ -47,7 +48,7 @@ import Darcs.Repository.Diff( treeDiff )
 import Darcs.Patch ( RepoPatch, PrimOf, PrimPatch, adddir, rmdir, addfile, rmfile,
                      listTouchedFiles )
 import Darcs.Patch.Apply( ApplyState )
-import Darcs.Patch.Witnesses.Ordered ( FL(..), (+>+), nullFL )
+import Darcs.Patch.Witnesses.Ordered ( FL(..), concatGapsFL, nullFL )
 import Darcs.Patch.Witnesses.Sealed ( Sealed(..), Gap(..), FreeLeft, unFreeLeft )
 import Darcs.Repository.Prefs ( filetypeFunction, FileType )
 import Darcs.Util.Tree( Tree, TreeItem(..), explodePaths )
@@ -81,14 +82,11 @@ remove = DarcsCommand
     , commandPrereq = amInHashedRepository
     , commandCompleteArgs = knownFileArgs
     , commandArgdefaults = nodefaults
-    , commandAdvancedOptions = odesc removeAdvancedOpts
-    , commandBasicOptions = odesc removeBasicOpts
-    , commandDefaults = defaultFlags removeOpts
-    , commandCheckOptions = ocheck removeOpts
+    , commandOptions = removeOpts
     }
   where
     removeBasicOpts = O.repoDir ^ O.recursive
-    removeAdvancedOpts = O.useIndex ^ O.umask
+    removeAdvancedOpts = O.umask
     removeOpts = removeBasicOpts `withStdOpts` removeAdvancedOpts
 
 removeCmd :: (AbsolutePath, AbsolutePath) -> [DarcsFlag] -> [String] -> IO ()
@@ -98,18 +96,19 @@ removeCmd fps opts relargs = do
     paths <- pathsFromArgs fps relargs
     when (any isRoot paths) $
         fail "Cannot remove a repository's root directory!"
-    withRepoLock (dryRun ? opts) (useCache ? opts) YesUpdatePending (umask ? opts) $
+    withRepoLock (useCache ? opts) (umask ? opts) $
       RepoJob $ \repository -> do
-        recorded_and_pending <- readRecordedAndPending repository
+        recorded_and_pending <- readPristineAndPending repository
         let exploded_paths =
               (if parseFlags O.recursive opts
                 then reverse . explodePaths recorded_and_pending
                 else id) paths
         Sealed p <- makeRemovePatch opts repository exploded_paths
-        -- TODO whether command fails depends on verbosity BAD BAD BAD
-        when (nullFL p && not (null paths) && not (quiet opts)) $
+        when (nullFL p && not (null paths)) $
             fail "No files were removed."
-        addToPending repository (O.useIndex ? opts) p
+        addToPending repository (diffingOpts opts) p
+        void $ finalizeRepositoryChanges repository
+            (O.dryRun ? opts)
         putInfo opts $ vcat $ map text $ ["Will stop tracking:"] ++
             map displayPath (listTouchedFiles p)
 
@@ -118,26 +117,29 @@ removeCmd fps opts relargs = do
 --   flag should be handled by the caller by adding all offspring of a directory
 --   to the files list.
 makeRemovePatch :: (RepoPatch p, ApplyState p ~ Tree)
-                => [DarcsFlag] -> Repository rt p wR wU wR
+                => [DarcsFlag] -> Repository rt p wU wR
                 -> [AnchoredPath] -> IO (Sealed (FL (PrimOf p) wU))
-makeRemovePatch opts repository files =
-                          do recorded <- T.expand =<< readRecordedAndPending repository
-                             unrecorded <- readUnrecorded repository (O.useIndex ? opts) $ Just files
-                             ftf <- filetypeFunction
-                             result <- foldM removeOnePath (ftf,recorded,unrecorded, []) files
-                             case result of
-                                 (_, _, _, patches) -> return $
-                                                         unFreeLeft $ foldr (joinGap (+>+)) (emptyGap NilFL) $ reverse patches
-    where removeOnePath (ftf, recorded, unrecorded, patches) f = do
-            let recorded' = T.modifyTree recorded f Nothing
-                unrecorded' = T.modifyTree unrecorded f Nothing
-            local <- makeRemoveGap opts ftf recorded unrecorded unrecorded' f
-            -- we can tell if the remove succeeded by looking if local is
-            -- empty. If the remove succeeded, we should pass on updated
-            -- recorded and unrecorded that reflect the removal
-            return $ case local of
-                       Just gap -> (ftf, recorded', unrecorded', gap : patches)
-                       _        -> (ftf, recorded, unrecorded, patches)
+makeRemovePatch opts repository files = do
+  recorded <- T.expand =<< readPristineAndPending repository
+  unrecorded <- readUnrecorded repository (O.useIndex ? opts) $ Just files
+  ftf <- filetypeFunction
+  result <- foldM removeOnePath (ftf, recorded, unrecorded, []) files
+  case result of
+    (_, _, _, patches) ->
+      return $
+      unFreeLeft $ concatGapsFL $ reverse patches
+  where
+    removeOnePath (ftf, recorded, unrecorded, patches) f = do
+      let recorded' = T.modifyTree recorded f Nothing
+          unrecorded' = T.modifyTree unrecorded f Nothing
+      local <- makeRemoveGap opts ftf recorded unrecorded unrecorded' f
+      -- we can tell if the remove succeeded by looking if local is
+      -- empty. If the remove succeeded, we should pass on updated
+      -- recorded and unrecorded that reflect the removal
+      return $
+        case local of
+          Just gap -> (ftf, recorded', unrecorded', gap : patches)
+          _ -> (ftf, recorded, unrecorded, patches)
 
 -- | Takes a file path and returns the FL of patches to remove that, wrapped in
 --   a 'Gap'.
